from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import List, Optional, Union, cast

import numpy as np
import torch.utils.tensorboard

from ... import _deprecation
from .._forward_declarations import _BuddyForwardDeclarations
from ._optimizer import _BuddyOptimizer


@dataclass
class _LogNamespace:
    buddy: _BuddyLogging
    scope: str

    def __enter__(self):
        self.buddy.log_scope_push(self.scope)
        return self

    def __exit__(self, *unused):
        self.buddy.log_scope_pop(self.scope)
        return


class _BuddyLogging(_BuddyForwardDeclarations, abc.ABC):
    """Buddy's TensorBoard logging interface.
    """

    def __init__(self, log_dir: str) -> None:
        """Logging-specific setup.

        Args:
            log_dir (str): Path to save Tensorboard logs to.
        """
        self._log_dir = log_dir

        # Backwards-compatibility for deprecated API
        self.log = _deprecation.new_name_wrapper(
            "Buddy.log()", "Buddy.log_scalar()", self.log_scalar
        )

        # State variables for TensorBoard
        # Note that the writer is lazily instantiated; see below
        self._log_writer: Optional[torch.utils.tensorboard.SummaryWriter] = None
        self._log_scopes: List[str] = []

    def log_scope(self, scope: str) -> _LogNamespace:
        """Returns a scope to log tensors in.

        Example usage:

        ```
            with buddy.log_scope("scope"):
                # Logs to scope/loss
                buddy.log_scalar("loss", loss_tensor)
        ```

        Args:
            scope (str): Name of scope.

        Returns:
            _LogNamespace: Object for automatically pushing/popping scope.
        """
        return _LogNamespace(self, scope)

    def log_scope_push(self, scope: str) -> None:
        """Push a scope to log tensors into.

        Example usage:
        ```
            buddy.log_scope_push("scope")

            # Logs to scope/loss
            buddy.log_scalar("loss", loss_tensor)

            buddy.log_scope_pop("scope") # name parameter is optional

        Args:
            scope (str): Name of scope.
        ```
        """
        self._log_scopes.append(scope)

    def log_scope_pop(self, scope: str = None) -> None:
        """Pop a scope we logged tensors into. See `log_scope_push()`.

        Args:
            scope (str, optional): Name of scope. Needs to be the top one in the stack.
        """
        popped = self._log_scopes.pop()
        if scope is not None:
            assert popped == scope, f"{popped} does not match {scope}!"

    def log_image(
        self,
        name: str,
        image: Union[torch.Tensor, np.ndarray],
        dataformats: str = "CHW",
    ) -> None:
        """Convenience function for logging an image tensor for visualization in
        TensorBoard.

        Equivalent to:
        ```
        buddy.log_writer.add_image(
            buddy.log_scope_prefix(name),
            image,
            buddy.optimizer_steps,
            dataformats
        )
        ```

        Args:
            name (str): Identifier for Tensorboard.
            image (torch.Tensor or np.ndarray): Image to log.
            dataformats (str, optional): Dimension ordering. Defaults to "CHW".
        """
        # Add scope prefixes
        name = self.log_scope_prefix(name)

        # Log scalar
        optimizer_steps = cast(_BuddyOptimizer, self).optimizer_steps
        self.log_writer.add_image(
            name, image, global_step=optimizer_steps, dataformats=dataformats
        )

    def log_scalar(
        self, name: str, value: Union[torch.Tensor, np.ndarray, float]
    ) -> None:
        """Convenience function for logging a scalar for visualization in TensorBoard.

        Equivalent to:
        ```
        buddy.log_writer.add_scalar(
            buddy.log_scope_prefix(name),
            value,
            buddy.optimizer_steps
        )
        ```

        Args:
            name (str): Identifier for Tensorboard.
            value (torch.Tensor, np.ndarray, or float): Value to log.
        """
        # Add scope prefixes
        name = self.log_scope_prefix(name)

        # Log scalar
        optimizer_steps = cast(_BuddyOptimizer, self).optimizer_steps
        self.log_writer.add_scalar(name, value, global_step=optimizer_steps)

    def log_scope_prefix(self, name: str = "") -> str:
        """Get or apply the current log scope prefix.

        Example usage:
        ```
        print(buddy.log_scope_prefix()) # ""

        with buddy.log_scope("scope0"):
            print(buddy.log_scope_prefix("loss")) # "scope0/loss"

            with buddy.log_scope("scope1"):
                print(buddy.log_scope_prefix()) # "scope0/scope1/"
        ```

        Args:
            name (str, optional): Name to prepend a prefix to. Defaults to an empty string.

        Returns:
            str: Scoped log name, or scope prefix if input is empty.
        """
        if len(self._log_scopes) == 0:
            return name
        else:
            return "{}/{}".format("/".join(self._log_scopes), name)

    @property
    def log_writer(self) -> torch.utils.tensorboard.SummaryWriter:
        """Accessor for standard Tensorboard SummaryWriter. Instantiated lazily.
        """
        if self._log_writer is None:
            self._log_writer = torch.utils.tensorboard.SummaryWriter(
                self._log_dir + "/" + self._experiment_name
            )
        return self._log_writer



from __future__ import annotations

import abc
import time
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Union, cast

import torch

from .._forward_declarations import _BuddyForwardDeclarations

if TYPE_CHECKING:
    from ._checkpointing import _BuddyCheckpointing


class _BuddyOptimizer(_BuddyForwardDeclarations, abc.ABC):
    """Buddy's optimization interface.
    """

    # Supported optimizer types
    # Note that torch (as of 1.5) has some stub issues with optim.Adadelta,
    # optim.Optimizer
    _OPTIMIZER_TYPES: Dict[str, torch.optim.Optimizer] = {  # type: ignore
        "adam": torch.optim.Adam,
        "adadelta": torch.optim.Adadelta,  # type: ignore
    }

    # Default learning rates
    _OPTIMIZER_DEFAULT_LEARNING_RATES: Dict[str, float] = {
        "adam": 1e-4,
        "adadelta": 1,
    }

    def __init__(
        self, optimizer_type: str, optimizer_checkpoint_interval: float
    ) -> None:
        """Optimizer-specific setup.
        """
        # Assign our training configuration.
        self._optimizer_config: Dict[str, Any] = {
            "global_steps": 0,
            "optimizer_type": optimizer_type,
            "learning_rate_schedulers": {},
        }

        # Map from optimizer name to optimizers
        # These are constructed lazily!
        self._optimizer_dict: Dict[str, torch.optim.Optimizer] = {}  # type: ignore

        # Autocheckpoint variables
        self._optimizer_checkpoint_interval: float = optimizer_checkpoint_interval
        self._optimizer_last_checkpoint_time: Optional[float] = None

        # Default learning rate
        self._optimizer_default_learning_rate: Union[
            float, Callable[[int], float]
        ] = self._OPTIMIZER_DEFAULT_LEARNING_RATES[optimizer_type]

    def minimize(
        self,
        loss: torch.Tensor,
        optimizer_name: str = "primary",
        *,
        retain_graph: bool = False,
        checkpoint_interval: float = None,
    ) -> None:
        """Compute gradients and use them to minimize a loss function.
        """
        assert self._model is not None, "No model attached!"
        self._instantiate_optimizer(optimizer_name)

        # Update learning rate using scheduler if possible
        schedulers = self._optimizer_config["learning_rate_schedulers"]
        if optimizer_name in schedulers:
            self._set_learning_rate(
                schedulers[optimizer_name](self._optimizer_config["global_steps"]),
                optimizer_name,
            )

        # Take gradient step
        self._optimizer_dict[optimizer_name].zero_grad()
        loss.backward(retain_graph=retain_graph)  # type: ignore
        self._optimizer_dict[optimizer_name].step()

        # Update global step count
        self._optimizer_config["global_steps"] += 1

        # Autocheckpoint procedure
        if checkpoint_interval is None:
            checkpoint_interval = self._optimizer_checkpoint_interval

        # Disable autocheckpoint if interval is 0
        if checkpoint_interval == 0:
            return

        if self._optimizer_last_checkpoint_time is None:
            # First iteration
            self._optimizer_last_checkpoint_time = time.time()
        elif (
            time.time() - cast(float, self._optimizer_last_checkpoint_time)
            > self._optimizer_checkpoint_interval
        ):  # pragma: no cover
            # Checkpoint!
            cast("_BuddyCheckpointing", self).save_checkpoint()
            self._optimizer_last_checkpoint_time = time.time()

    def get_learning_rate(self, optimizer_name: str = "primary") -> float:
        """Gets an optimizer learning rate.
        """
        assert self._model is not None, "No model attached!"
        assert optimizer_name in self._optimizer_dict

        # Return scheduled learning rate
        schedulers = self._optimizer_config["learning_rate_schedulers"]
        if optimizer_name in schedulers:
            return schedulers[optimizer_name](self.optimizer_steps)

        # Return raw learning rate
        # Currently, only one parameter group is supported
        optimizer = self._optimizer_dict[optimizer_name]
        assert len(optimizer.param_groups) == 1
        return optimizer.param_groups[0]["lr"]

    def set_learning_rate(
        self,
        value: Union[float, Callable[[int], float]],
        optimizer_name: str = "primary",
    ) -> None:
        """Sets an optimizer learning rate. Accepts either a floating point
        learning rate or a schedule function (int steps -> float LR).
        """
        assert self._model is not None, "No model attached!"
        schedulers = self._optimizer_config["learning_rate_schedulers"]
        if callable(value):
            # Store a scheduler
            assert type(value(0)) == float
            schedulers[optimizer_name] = value
        else:
            # Set learning rate to a float
            assert type(value) == float
            # Delete scheduler
            if optimizer_name in schedulers.keys():
                schedulers.pop(optimizer_name)

            # Set scalar learning rate
            self._set_learning_rate(value, optimizer_name)

    def set_default_learning_rate(
        self, value: Union[float, Callable[[int], float]]
    ) -> None:
        """Sets a default learning rate for new optimizers.
        """
        self._optimizer_default_learning_rate = value

    @property
    def optimizer_steps(self) -> int:
        """Read-only interface for # of steps taken by optimizer.
        """
        return self._optimizer_config["global_steps"]

    def _set_learning_rate(self, value: float, optimizer_name: str) -> None:
        """(Private) Sets an optimizer's learning rate.
        """

        self._instantiate_optimizer(optimizer_name)

        # Currently, only one parameter group is supported
        optimizer = self._optimizer_dict[optimizer_name]
        assert len(optimizer.param_groups) == 1
        optimizer.param_groups[0]["lr"] = value

    def _instantiate_optimizer(self, optimizer_name: str) -> None:
        """(Private) Instantiates an optimizer. Returns immediately if
        optimizer already exists.
        """
        assert self._model is not None, "No model attached!"
        if optimizer_name in self._optimizer_dict.keys():
            # Optimizer already exists: do nothing!
            return

        self._print("Instantiating optimizer: ", optimizer_name)

        # Make sure we're creating a valid optimizer
        optimizer_type = self._optimizer_config["optimizer_type"]
        assert optimizer_type in self._OPTIMIZER_TYPES

        # Parameters
        Optimizer = self._OPTIMIZER_TYPES[optimizer_type]
        initial_learning_rate = self._optimizer_default_learning_rate

        # Construct optimizer
        self._optimizer_dict[optimizer_name] = Optimizer(
            self._model.parameters(), lr=initial_learning_rate
        )



import abc

import torch
import torch.nn as nn


class Base(nn.Module, abc.ABC):
    def __init__(self, activation: str = "relu", activations_inplace: bool = False):
        super().__init__()

        self.activations_inplace = activations_inplace

        self.block1: nn.Module
        self.block2: nn.Module
        self.activation = self._activation_func(activation)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """ResBlock forward pass.
        """
        residual = x
        x = self.block1(x)
        x = self.activation(x)
        x = self.block2(x)
        assert x.shape[0] == residual.shape[0]
        x += residual
        x = self.activation(x)
        return x

    def _activation_func(self, activation: str) -> nn.Module:
        return nn.ModuleDict(
            {
                "relu": nn.ReLU(inplace=self.activations_inplace),
                "leaky_relu": nn.LeakyReLU(inplace=self.activations_inplace),
                "selu": nn.SELU(inplace=self.activations_inplace),
                "none": nn.Identity(),
            }
        )[activation]


class Linear(Base):
    def __init__(self, units: int, bottleneck_units: int = None, **resblock_base_args):
        super().__init__(**resblock_base_args)

        if bottleneck_units is None:
            bottleneck_units = units
        self.block1 = nn.Linear(units, bottleneck_units)
        self.block2 = nn.Linear(bottleneck_units, units)


class Conv2d(Base):
    def __init__(
        self,
        channels: int,
        bottleneck_channels: int = None,
        kernel_size: int = 3,
        **resblock_base_args
    ):
        super().__init__(**resblock_base_args)

        if bottleneck_channels is None:
            bottleneck_channels = channels

        self.block1 = nn.Conv2d(
            channels,
            bottleneck_channels,
            kernel_size=kernel_size,
            padding=kernel_size // 2,
        )
        self.block2 = nn.Conv2d(
            bottleneck_channels,
            channels,
            kernel_size=kernel_size,
            padding=kernel_size // 2,
        )



import os

import numpy as np
import pytest
import torch
import torch.nn.functional as F

import fannypack

from ..fixtures import resblock_buddy_temporary_data, simple_buddy


def test_buddy_device(simple_buddy):
    """Make sure Buddy can list the existing checkpoint labels.
    """
    model, buddy, data, labels = simple_buddy
    assert isinstance(buddy.device, torch.device)


def test_buddy_checkpoint_labels(simple_buddy):
    """Make sure Buddy can list the existing checkpoint labels.
    """
    model, buddy, data, labels = simple_buddy
    assert set(buddy.checkpoint_labels) == set(["0000000000000200", "new", "legacy"])


def test_buddy_load_checkpoint_new_format_label(simple_buddy):
    """Make sure Buddy can load checkpoints via a label specifier.
    """
    model, buddy, data, labels = simple_buddy
    initial_loss = F.mse_loss(model(data), labels)

    buddy.load_checkpoint(label="new")
    final_loss = F.mse_loss(model(data), labels)

    assert final_loss < initial_loss / 2.0
    assert buddy.optimizer_steps == 200


def test_buddy_load_checkpoint_new_format_path(simple_buddy):
    """Make sure Buddy can load checkpoints directly by path.
    """
    model, buddy, data, labels = simple_buddy
    initial_loss = F.mse_loss(model(data), labels)

    buddy.load_checkpoint(
        path=os.path.join(
            # Use path relative to this test
            os.path.dirname(__file__),
            "../assets/checkpoints/simple_net-new.ckpt",
        )
    )
    final_loss = F.mse_loss(model(data), labels)

    assert final_loss < initial_loss / 2.0
    assert buddy.optimizer_steps == 200


def test_buddy_load_checkpoint_legacy_format_label(simple_buddy):
    """Make sure Buddy is backward-compatible with an old checkpoint
    format when we load it using a label specifier.
    """
    model, buddy, data, labels = simple_buddy
    initial_loss = F.mse_loss(model(data), labels)

    buddy.load_checkpoint(label="legacy")
    final_loss = F.mse_loss(model(data), labels)

    assert final_loss < initial_loss / 2.0
    assert buddy.optimizer_steps == 200


def test_buddy_load_checkpoint_legacy_format_path(simple_buddy):
    """Make sure Buddy is backward-compatible with an old checkpoint
    format when we load it directly by path.
    """
    model, buddy, data, labels = simple_buddy
    initial_loss = F.mse_loss(model(data), labels)

    buddy.load_checkpoint(
        path=os.path.join(
            # Use path relative to this test
            os.path.dirname(__file__),
            "../assets/checkpoints/simple_net-legacy.ckpt",
        )
    )
    final_loss = F.mse_loss(model(data), labels)

    assert final_loss < initial_loss / 2.0
    assert buddy.optimizer_steps == 200


def test_buddy_load_missing_checkpoint_path(simple_buddy):
    """Make sure Buddy raises an error if we load a nonexistent checkpoint.
    """
    model, buddy, data, labels = simple_buddy

    with pytest.raises(FileNotFoundError):
        buddy.load_checkpoint(
            path=os.path.join(
                os.path.dirname(__file__), "checkpoints/this_file_doesnt_exist.ckpt",
            )
        )


def test_buddy_load_missing_checkpoint_label(simple_buddy):
    """Make sure Buddy raises an error if we load a nonexistent checkpoint.
    """
    model, buddy, data, labels = simple_buddy

    with pytest.raises(FileNotFoundError):
        buddy.load_checkpoint(label="this_checkpoint_label_doesnt_exist")


def test_buddy_load_missing_checkpoint(simple_buddy):
    """Make sure Buddy raises an error if we load a nonexistent checkpoint.
    """
    model, buddy, data, labels = simple_buddy

    with pytest.raises(FileNotFoundError):
        buddy.load_checkpoint(experiment_name="invalid_experiment")


def test_buddy_load_checkpoint_experiment_name(simple_buddy):
    """Make sure Buddy can save/load a checkpoint.
    """
    model, buddy, data, labels = simple_buddy

    buddy.load_checkpoint(experiment_name="simple_net")
    assert buddy.optimizer_steps == 200


def test_buddy_save_checkpoint(resblock_buddy_temporary_data):
    """Make sure Buddy can save/load a checkpoint.
    """
    model, buddy, data, labels = resblock_buddy_temporary_data

    # Save initial state
    assert buddy.optimizer_steps == 0
    buddy.save_checkpoint()

    # Step once via ADAM
    loss = F.mse_loss(model(data), labels) * 1000
    buddy.minimize(loss)
    assert buddy.optimizer_steps == 1

    # Restore initial state and verify
    buddy.load_checkpoint()
    assert buddy.optimizer_steps == 0


def test_buddy_save_checkpoint_label(resblock_buddy_temporary_data):
    """Make sure Buddy can save/load a checkpoint with a label specifier.
    """
    model, buddy, data, labels = resblock_buddy_temporary_data

    # Save initial state
    assert buddy.optimizer_steps == 0
    buddy.save_checkpoint(label="memorable_name")

    # Step once via ADAM
    loss = F.mse_loss(model(data), labels)
    buddy.minimize(loss)
    assert buddy.optimizer_steps == 1

    # Save current state
    buddy.save_checkpoint()

    # Restore initial state and verify
    buddy.load_checkpoint(label="memorable_name")
    assert buddy.optimizer_steps == 0


def test_buddy_save_checkpoint_labels(resblock_buddy_temporary_data):
    """Make sure Buddy's list of existing checkpoints is correctly updated.
    """
    model, buddy, data, labels = resblock_buddy_temporary_data

    # No initial checkpoints
    assert len(buddy.checkpoint_labels) == 0

    # Save initial state
    buddy.save_checkpoint()

    # Check checkpoint_labels property
    assert len(buddy.checkpoint_labels) == 1
    assert int(buddy.checkpoint_labels[0]) == buddy.optimizer_steps

    # Save labeled state
    buddy.save_checkpoint(label="memorable_name")

    # Check checkpoint_labels property
    assert len(buddy.checkpoint_labels) == 2
    assert "memorable_name" in buddy.checkpoint_labels


def test_buddy_checkpoint_modules(resblock_buddy_temporary_data):
    """Test that Buddy can load individual modules from checkpoint files.
    """
    model, buddy, data, labels = resblock_buddy_temporary_data

    # Check that layer parameters are different
    layer2_params = dict(model.layer2.named_parameters())
    layer4_params = dict(model.layer4.named_parameters())
    assert set(layer2_params.keys()) == set(layer4_params.keys())
    for key in layer2_params.keys():
        if "bias" in key:
            continue
        assert not torch.allclose(layer2_params[key].data, layer4_params[key].data)

    # Save a checkpoint
    buddy.save_checkpoint()

    # Load module parameters from our checkpoint file
    buddy.load_checkpoint_module(source="layer2", target="layer4")

    # If we pass in just a single argument, we should be able to infer the target module
    # Note that this is loading from a checkpoint we just saved, so this operation won't
    # actually modify any parameters
    buddy.load_checkpoint_module("layer2")

    # Check that layer parameters are now identical
    layer2_params = dict(model.layer2.named_parameters())
    layer4_params = dict(model.layer4.named_parameters())
    assert set(layer2_params.keys()) == set(layer4_params.keys())
    for key in layer2_params.keys():
        if "bias" in key:
            continue
        assert torch.allclose(layer2_params[key].data, layer4_params[key].data)


def test_buddy_checkpoint_optimizers(resblock_buddy_temporary_data):
    """Sanity-check for Buddy's optimizer loading interface.
    """
    model, buddy, data, labels = resblock_buddy_temporary_data

    # Create some optimizer parameters
    buddy.set_learning_rate(1e-3)
    assert buddy.get_learning_rate() == 1e-3

    # Save a checkpoint
    buddy.save_checkpoint()

    # Overwrite optimizer parameter
    buddy.set_learning_rate(1e-5)
    assert buddy.get_learning_rate() == 1e-5
    # Restore optimizer parameters
    buddy.load_checkpoint_optimizers()
    assert buddy.get_learning_rate() == 1e-3

    # Overwrite optimizer parameter
    buddy.set_learning_rate(1e-5)
    assert buddy.get_learning_rate() == 1e-5
    # Restore optimizer parameters
    buddy.load_checkpoint_optimizer("primary")
    assert buddy.get_learning_rate() == 1e-3

    # Overwrite optimizer parameter
    buddy.set_learning_rate(1e-5)
    assert buddy.get_learning_rate() == 1e-5
    # Restore optimizer parameters
    buddy.load_checkpoint_optimizer(source="primary", target="primary")
    assert buddy.get_learning_rate() == 1e-3

    # Load nonexistent optimizer parameter: these should fail
    with pytest.raises(AssertionError):
        buddy.load_checkpoint_optimizer(
            source="this_optimizer_does_not_exist", target="primary"
        )
    with pytest.raises(AssertionError):
        buddy.load_checkpoint_optimizer("this_optimizer_does_not_exist")


def test_buddy_checkpointed_train(resblock_buddy_temporary_data):
    """Do some checks on rapid continuous checkpointing.
    """
    model, buddy, data, labels = resblock_buddy_temporary_data
    model.train()

    # Save an initial checkpoint
    buddy.save_checkpoint()

    # This should do nothing
    buddy.save_checkpoint()

    assert len(buddy.checkpoint_labels) == 1
    orig_checkpoint = buddy.checkpoint_labels[0]

    # Do some training
    initial_loss = F.mse_loss(model(data), labels)
    buddy.set_learning_rate(1e-3)
    for _ in range(10):
        loss = F.mse_loss(model(data), labels)
        buddy.minimize(loss)
        buddy.save_checkpoint()

    # Checkpoint count should be truncated
    assert len(buddy.checkpoint_labels) == 5
    assert orig_checkpoint not in buddy.checkpoint_labels

    # Loss should at least have halved
    final_loss = F.mse_loss(model(data), labels)
    assert final_loss < initial_loss / 2.0



import pytest
import torch

import fannypack

from ..fixtures import simple_net


def test_freeze_unfreeze(simple_net):
    """Check to make sure freezing/unfreezing a module behaves as expected.
    """

    # All parameters should be initially unfrozen
    _check_frozen(simple_net, False)

    # Freeze the module
    fannypack.utils.freeze_module(simple_net)

    # All parameters should be frozen
    _check_frozen(simple_net, True)

    # Unfreeze the module
    fannypack.utils.unfreeze_module(simple_net)

    # All parameters should be unfrozen
    _check_frozen(simple_net, False)


def test_unfreeze_children(simple_net):
    """Check to make sure unfreezing child modules behaves as expected.
    """

    # All parameters should be initially unfrozen
    _check_frozen(simple_net, False)

    # Freeze the module
    fannypack.utils.freeze_module(simple_net)

    # All parameters should be frozen
    _check_frozen(simple_net, True)

    # Unfreeze the module
    fannypack.utils.unfreeze_module(simple_net.layer1)
    fannypack.utils.unfreeze_module(simple_net.layer2)
    fannypack.utils.unfreeze_module(simple_net.layer3)
    fannypack.utils.unfreeze_module(simple_net.layer4)

    # All parameters should be unfrozen
    _check_frozen(simple_net, False)


def test_freeze_children(simple_net):
    """Check to make sure freezing child modules behaves as expected.
    """

    # All parameters should be initially unfrozen
    _check_frozen(simple_net, False)

    # Freeze a child module
    fannypack.utils.freeze_module(simple_net.layer1)

    # All parameters for layer1 should be frozen
    _check_frozen(simple_net.layer1, True)
    _check_frozen(simple_net.layer2, False)
    _check_frozen(simple_net.layer3, False)
    _check_frozen(simple_net.layer4, False)

    # Unfreeze the entire module
    fannypack.utils.unfreeze_module(simple_net)

    # All parameters should be unfrozen
    _check_frozen(simple_net, False)


# Helper to check whether a module is frozen or not
def _check_frozen(module, is_frozen):
    for param in module.parameters():
        assert param.requires_grad == (not is_frozen)



import abc
import os
from typing import Any, Dict

import yaml

from .._forward_declarations import _BuddyForwardDeclarations


class _BuddyMetadata(_BuddyForwardDeclarations, abc.ABC):
    """Buddy's experiment metadata management interface.
    """

    def __init__(self, metadata_dir: str) -> None:
        """Metadata-specific setup.
        """

        # Attempt to read existing metadata
        self._metadata_dir = metadata_dir
        try:
            self.load_metadata()
        except FileNotFoundError:
            self._metadata: Dict[str, Any] = {}

    def load_metadata(
        self, experiment_name: str = None, metadata_dir: str = None, path: str = None
    ) -> None:
        """Read existing metadata file.
        """
        if path is None:
            if experiment_name is None:
                experiment_name = self._experiment_name
            if metadata_dir is None:
                metadata_dir = self._metadata_dir
            path = os.path.join(metadata_dir, f"{experiment_name}.yaml")
        else:
            assert experiment_name is None and metadata_dir is None

        with open(path, "r") as file:
            self._metadata = yaml.load(file, Loader=yaml.SafeLoader)
            self._print("Loaded metadata:", self._metadata)

    def add_metadata(self, content: Dict[str, Any]) -> None:
        """Add human-readable metadata for this experiment. Input should be a
        dictionary that is merged with existing metadata.
        """
        assert type(content) == dict

        # Merge input metadata with current metadata
        for key, value in content.items():
            self._metadata[key] = value

        # Write to disk
        self._write_metadata()

    def set_metadata(self, content: Dict[str, Any]) -> None:
        """Assign human-readable metadata for this experiment. Input should be
        a dictionary that replaces existing metadata.
        """
        assert type(content) == dict

        # Replace input metadata with current metadata
        self._metadata = content

        # Write to disk
        self._write_metadata()

    def _write_metadata(self) -> None:
        # Create metadata directory if needed
        if not os.path.isdir(self._metadata_dir):
            os.makedirs(self._metadata_dir)
            self._print("Created directory:", self._metadata_dir)

        # Write metadata to file
        metadata_path = "{}/{}.yaml".format(self._metadata_dir, self._experiment_name,)
        with open(metadata_path, "w") as file:
            yaml.dump(self._metadata, file, default_flow_style=False)
            self._print("Wrote metadata to:", metadata_path)

    @property
    def metadata(self) -> Dict[str, Any]:
        """Read-only interface for experiment metadata.
        """
        return self._metadata

    @property
    def metadata_path(self) -> str:
        """Read-only path to my metadata file.
        """
        return os.path.join(self._metadata_dir, f"{self._experiment_name}.yaml")



import dataclasses
import functools
from typing import Any, Callable, Dict, List, Tuple, TypeVar, Union, cast, overload

import numpy as np
import torch

# Type aliases, variables
# > We'd ideally be able to use a TypeVar for Container, but this causes an overlapping
#   function signatures error in mypy
Container = Any
InputType = TypeVar("InputType", np.ndarray, torch.Tensor)
OutputType = TypeVar("OutputType", np.ndarray, torch.Tensor)


@overload
def _convert_recursive(
    x: InputType, convert: Callable[[InputType], OutputType], input_type: type,
) -> OutputType:
    pass


@overload
def _convert_recursive(
    x: Container, convert: Callable[[InputType], OutputType], input_type: type,
) -> Container:
    pass


def _convert_recursive(x, convert, input_type):
    """Private conversion helper. Recursively calls a conversion function on inputs
    within a nested set of containers.
    """

    # Conversion base case
    if type(x) == input_type:
        x = cast(InputType, x)
        return convert(x)

    # Convert containers: bind arguments to helper function
    convert_recursive = functools.partial(
        _convert_recursive, convert=convert, input_type=input_type
    )

    # Convert dictionaries of values
    if type(x) == dict:
        x = cast(dict, x)
        return {k: convert_recursive(v) for k, v in x.items()}

    # Convert lists of values
    elif type(x) == list:
        x = cast(list, x)
        return list(map(convert_recursive, x))

    # Convert tuples of values
    elif type(x) == tuple:
        x = cast(tuple, x)
        return tuple(map(convert_recursive, x))

    # Convert dataclass containing values
    elif dataclasses.is_dataclass(x):
        changes = {}
        for field in dataclasses.fields(x):
            value = getattr(x, field.name)
            try:
                changes[field.name] = convert_recursive(value)
            except AssertionError as e:
                # For dataclasses, we leave unsupported types alone
                # May want to rethink this?
                pass
        return dataclasses.replace(x, **changes)

    # Unsupported input types
    else:
        assert False, f"Unsupported datatype {type(x)}!"


@overload
def to_device(
    x: torch.Tensor, device: torch.device, detach: bool = False
) -> torch.Tensor:
    pass


@overload
def to_device(x: Container, device: torch.device, detach: bool = False) -> Container:
    pass


def to_device(
    x: Union[Container, torch.Tensor], device: torch.device, detach: bool = False
) -> Union[Container, torch.Tensor]:
    """Move a torch tensor, list of tensors, dict, or dataclass of tensors to a
    different device.

    Args:
        x: (torch.Tensor, list, tuple, dict, or dataclass) Tensor or container of
            tensors to move.
        device (torch.device): Target device.
        detach (bool, optional): If set, detaches tensors after moving. Defaults to
            False.

    Returns:
        torch.Tensor, list, tuple, dict, or dataclass: Output, type will mirror input.
    """

    def convert(x):
        if detach:
            x = x.detach()
        return x.to(device)

    return _convert_recursive(x, convert=convert, input_type=torch.Tensor)


@overload
def to_torch(
    x: np.ndarray, device: str = "cpu", convert_doubles_to_floats: bool = True,
) -> torch.Tensor:
    pass


@overload
def to_torch(
    x: Container, device: str = "cpu", convert_doubles_to_floats: bool = True,
) -> Container:
    pass


def to_torch(
    x, device="cpu", convert_doubles_to_floats=True,
):
    """Converts a numpy array, list of numpy arrays, dict, or dataclass of numpy arrays
    for use in PyTorch.

    Args:
        x: (np.ndarray, list, tuple, dict, or dataclass) Array or container of arrays to
            convert to torch tensors.
        device (torch.device, optional): Torch device to create tensors on. Defaults to
            `"cpu"`.
        convert_doubles_to_floats (bool, optional): If set, converts 64-bit floats to
            32-bit. Defaults to True.

    Returns:
        torch.Tensor, list, tuple, dict, or dataclass: Output, type will mirror input.
    """

    def convert(x: np.ndarray) -> torch.Tensor:
        output = torch.from_numpy(x)
        if x.dtype == np.float64 and convert_doubles_to_floats:
            output = output.float()
        output = output.to(device)
        return output

    return _convert_recursive(x, convert=convert, input_type=np.ndarray)


@overload
def to_numpy(x: torch.Tensor) -> np.ndarray:
    pass


@overload
def to_numpy(x: Container) -> Container:
    pass


def to_numpy(x):
    """Converts a tensor, list of tensors, dict, or dataclass of tensors for use in
    Numpy.

    Args:
        x: (torch.Tensor, list, tuple, dict, or dataclass) Tensor or container of
            tensors to convert to numpy.

    Returns:
        np.ndarray, list, tuple, dict, or dataclass: Output, type will mirror input.
    """

    def convert(x: torch.Tensor) -> np.ndarray:
        output = x.detach().cpu().numpy()
        return output

    return _convert_recursive(x, convert=convert, input_type=torch.Tensor)



from typing import Any, Dict, Optional, Union

import h5py
import numpy as np


class TrajectoriesFile:
    """An interface for reading/writing trajectory files using hdf5 files.

    Each TrajectoriesFile represents an iterable list of trajectories, where
    each trajectory is stored as a dictionary that maps `str` keys to
    `np.ndarray` contents.

    Example usage (read):
    ```
    with TrajectoriesFile('test.hdf5') as traj_file:

        for traj in traj_file:
            print(traj.keys()) # list of keys
            print(traj['some-key-name']) # numpy array
    ```

    Example usage (write):
    ```
    traj_file = TrajectoriesFile('test.hdf5', read_only=False)

    traj_file.add_meta({'label': 5})
    traj_file.add_timestep({'a': 1, 'b': 2})
    traj_file.add_timestep({'a': 3, 'b': 4})

    with traj_file:
        traj_file.complete_trajectory()

    print(len(traj_file)) # 1 trajectory!

    with traj_file:
        print(traj_file[0]['label']) # 5
        print(traj_file[0]['a']) # [1, 3]
        print(traj_file[0]['b']) # [2, 4]
    ```

    Note that some operations -- ones that require interfacing with the
    filesytem -- need to be called within a `with` statement.

    """

    def __init__(
        self,
        path: str,
        convert_doubles: bool = True,
        read_only: bool = True,
        compress: bool = True,
        verbose: bool = True,
    ):
        """Constructs an interface for reading from/writing to hdf5 files.

        Args:
            path (str): File path for this trajectory file.
            convert_doubles (bool): Convert doubles to floats to shrink files.
            read_only (bool, optional): Open file in read-only mode.
            compress (bool, optional): Reduce filesize w/ gzip.
            verbose (bool, optional): Enable debug prints.
        """
        assert path[-5:] == ".hdf5", "Missing file extension!"

        # Meta
        self._path: str = path
        self._convert_doubles: bool = convert_doubles
        self._read_only: bool = read_only
        self._compress: bool = compress
        self._verbose: bool = verbose

        # Maps content key => content
        self._content_dict: Dict[str, Any] = {}

        # Number of timesteps in current trajectory
        self._current_trajectory_timesteps: int = 0

        # Count the number of trajectories that already exist
        self._trajectory_prefix: str = "trajectory"
        with self._h5py_file() as f:
            self._print("Loading trajectory from file:", f)
            if len(f.keys()) > 0:
                prefix_length = len(self._trajectory_prefix)
                ids = [int(k[prefix_length:]) for k in f.keys()]
                self._trajectory_count = max(ids) + 1
            else:
                self._trajectory_count = 0

            self._print("Existing trajectory count:", self._trajectory_count)

        assert type(self._trajectory_count) == int

        # File object
        self._file: Optional[h5py.File] = None

    def __enter__(self):
        """Automatic file opening, for use in `with` statements.
        """
        if self._file is None:
            self._print("Opening file...")
            self._file = self._h5py_file()
        return self

    def __exit__(self, *unused):
        """Automatic file closing, for use in `with` statements.
        """
        if self._file is not None:
            self._print("Closing file...")
            self._file.close()
            self._file = None

    def __getitem__(self, index: int) -> Dict[str, Union[np.ndarray, str]]:
        """Accessor for individual trajectories held by this file.
        Must be called with the TrajectoriesFile object in a `with` statement.

        Args:
            index (int): Trajectory #.

        Returns:
            dict: A (str->value) map containing data collected at each timestep of our
            trajectory.
        """
        assert self._file is not None, "Not called in with statement!"

        # Index checks
        assert type(index) == int
        if index < 0 and index >= -len(self):
            # Negative indexing
            index = index % len(self)
        elif index >= len(self) or index < -len(self):
            # For use as a standard Python iterator
            raise IndexError

        traj_key = self._trajectory_prefix + str(index)
        assert traj_key in self._file.keys()

        # Copy values to numpy array
        output: Dict[str, Union[np.ndarray, str]] = {}
        for key, value in self._file[traj_key].items():
            # Conversion
            converted_value = np.array(value)

            # Numpy strings => native strings
            if converted_value.dtype.type is np.string_:
                # Decode
                output[key] = bytes(converted_value).decode("utf-8")
            else:
                output[key] = converted_value

        return output

    def __setitem__(self, index: int, item: Dict[str, Union[np.ndarray, str]]):
        """Assignment operation for modifying or mutating trajectories.
        Must be called with the TrajectoriesFile object in a `with` statement.

        Args:
            index (int): Trajectory #.
            item (dict): A (str->value) map, as would be returned by __getitem__().
        """
        assert self._file is not None, "Not called in with statement!"

        # Input type validation
        assert type(item) == dict

        # Index checks
        assert type(index) == int
        if index < 0 and index >= -len(self):
            # Negative indexing
            index = index % len(self)
        elif index >= len(self) or index < -len(self):
            # For use as a standard Python iterator
            raise IndexError

        traj_key = self._trajectory_prefix + str(index)
        group = self._file[traj_key]

        # Delete anything that's already in this trajectory
        for key in group:
            del group[key]

        # Populate the trajectory
        for key, value in item.items():
            # Convert content to a numpy array
            if type(value) == str:
                data = np.string_(value)
            else:
                data = np.asarray(value)

            # Compress floats
            if data.dtype == np.float64 and self._convert_doubles:
                data = data.astype(np.float32)

            # Disable chunking, compression for scalars
            if len(data.shape) == 0:
                group.create_dataset(key, data=data)
            else:
                group.create_dataset(
                    key,
                    data=data,
                    chunks=True,
                    compression="gzip" if self._compress else None,
                )

    def __len__(self) -> int:
        """Returns the number of recorded trajectories.
        """
        return self._trajectory_count

    def resize(self, count: int):
        """Expand or contract our TrajectoriesFile.
        """
        assert self._file is not None, "Not called in with statement!"

        if self._trajectory_count <= count:
            # Expand
            for index in range(self._trajectory_count, count):
                traj_key = self._trajectory_prefix + str(index)
                self._file.create_group(traj_key)

        elif self._trajectory_count > count:
            # Contract
            for index in range(count, self._trajectory_count):
                traj_key = self._trajectory_prefix + str(index)
                del self._file[traj_key]

        self._trajectory_count = count

    def get_all(self, key: str) -> list:
        """Get contents associated with a key from all trajectories.

        Args:
            key (str): Content identifier.
        Returns:
            list: List of contents. First index is trajectory #.
        """
        assert self._file is not None, "Not called in with statement!"

        # Copy values to numpy array
        output = []
        for index in range(self._trajectory_count):
            traj_key = self._trajectory_prefix + str(index)
            assert traj_key in self._file.keys()
            value = self._file[traj_key][key]

            # Conversion
            value = np.array(value)
            assert type(value) == np.ndarray

            # Numpy strings => native strings
            if value.dtype.type is np.string_:
                # Decode
                value = bytes(value).decode("utf-8")

            # Add to output
            output.append(value)
        return output

    def add_timestep(self, content: Dict[str, np.ndarray]) -> None:
        """Add a timestep to the current trajectory.

        Args:
            content (dict): Map from timestep keys (str) to values (np.ndarray).
        """
        for key, value in content.items():
            if key not in self._content_dict:
                self._content_dict[key] = []

            assert type(self._content_dict[key]) == list
            self._content_dict[key].append(np.copy(value))

        # Increment length
        self._current_trajectory_timesteps += 1

    def add_meta(self, content: Dict[str, np.ndarray]) -> None:
        """Add some metadata to the current trajectory.

        Args:
            content (dict): Map from metadata keys (str) to values (np.ndarray).
        """
        for key, value in content.items():
            assert key not in self._content_dict.keys()
            if type(value) == str:
                self._content_dict[key] = value
            else:
                self._content_dict[key] = np.copy(value)

    def abandon_trajectory(self) -> None:
        """Abandon the current trajectory.
        """
        self._print("Abandoning trajectory")
        self._content_dict = {}
        self._current_trajectory_timesteps = 0

    def complete_trajectory(self) -> None:
        """Write the current trajectory to disk, and mark the start of a new
        trajectory.
        Must be called with the TrajectoriesFile object in a `with` statement.

        The next call to `add_timestep()` will be time 0 of the next trajectory.
        """
        assert self._file is not None, "Not called in with statement!"

        if self._content_dict == {}:
            self._print("Empty trajectory; skipping complete_trajectory()")
            return

        # Check length, print debug message
        self._print(
            "Completing trajectory!" f"(length={self._current_trajectory_timesteps})"
        )

        # Make space for an extra trajectory
        self.resize(self._trajectory_count + 1)

        # Copy contents in!
        self[-1] = self._content_dict

        # Reset state
        self._content_dict = {}
        self._current_trajectory_timesteps = 0
        self._print("Existing trajectory count:", self._trajectory_count)

    def clear(self) -> None:
        """Clear the contents of the TrajectoriesFile.
        """
        assert self._file is not None, "Not called in with statement!"

        for traj_key in self._file.keys():
            del self._file[traj_key]

    def _h5py_file(self, mode: str = None) -> h5py.File:
        """Private helper for creating h5py file objects.
        """
        if mode is None:
            mode = "r" if self._read_only else "a"

        return h5py.File(self._path, mode)

    def _print(self, *args, **kwargs):
        """Private helper for logging.
        """
        # Only print in verbose mode
        if self._verbose:
            identifier = self._path
            if len(identifier) > 20:
                identifier = f"...{identifier[-17:]}"

            args = list(args)
            args[0] = f"[TrajectoriesFile-{identifier}] {args[0]}"
            print(*args, **kwargs)



import warnings
from typing import List, Optional, cast

import torch
import torch.nn as nn

from ._buddy_include._interfaces._checkpointing import _BuddyCheckpointing
from ._buddy_include._interfaces._logging import _BuddyLogging
from ._buddy_include._interfaces._metadata import _BuddyMetadata
from ._buddy_include._interfaces._optimizer import _BuddyOptimizer


class Buddy(
    _BuddyCheckpointing, _BuddyOptimizer, _BuddyLogging, _BuddyMetadata,
):

    """Buddy is a model manager that abstracts away PyTorch boilerplate.
    Buddy helps with...
    - Creating/using/managing optimizers,
    - Checkpointing (models + optimizers),
    - Namespaced/scoped Tensorboard logging,
    - Saving human-readable metadata files.

    Args:
        experiment_name (str): Name for the current model/experiment.
        model (torch.nn.Module): PyTorch model to work with.

    Keyword Args:
        checkpoint_dir (str, optional): Path to save checkpoints into.
            Defaults to `"checkpoints"`.
        checkpoint_max_to_keep (int, optional): Number of auto-saved
            checkpoints to keep. Set to `None` to keep all. Defaults to `5`.
        metadata_dir (str, optional): Path to save metadata YAML files into.
            Defaults to `"metadata"`.
        log_dir (str, optional): Path to save Tensorboard log files into.
            Defaults to `"logs"`.
        optimizer_type (str, optional): Optimizer type to use: `"adam"` or
            `"adadelta"`. Defaults to `"adam"`.
        optimizer_checkpoint_interval (float, optional): How often to
            auto-checkpoint, as an interval in seconds. Time is computed from
            the first call to minimize(). Set to 0 to disable. Defaults to 300.
        verbose (bool, optional): Flag for toggling debug messages. Defaults to
            `True`.
        cpu_only (bool, optional): Set to True to turn off auto-detection of cuda
            support.
    """

    def __init__(
        self,
        experiment_name: str,
        model: Optional[nn.Module] = None,
        *,
        checkpoint_dir: str = "checkpoints",
        checkpoint_max_to_keep: int = 5,
        metadata_dir: str = "metadata",
        log_dir: str = "logs",
        optimizer_type: str = "adam",
        optimizer_checkpoint_interval: float = 300,
        optimizer_names: Optional[List[str]] = None,  # Deprecated!
        verbose: bool = True,
        cpu_only: bool = False,
    ) -> None:
        """Constructor
        """
        # Validate and assign core parameters.
        assert type(experiment_name) == str
        assert type(verbose) == bool

        self._experiment_name = experiment_name
        self._verbose = True

        # Use GPU for training if available.
        if not cpu_only and torch.cuda.is_available():
            self._device = torch.device("cuda")
        else:
            self._device = torch.device("cpu")
        self._print("Using device:", self._device)

        # Attach model if available
        if model != None:
            self.attach_model(cast(nn.Module, model))
        else:
            self._model = None

        # Call constructors for each of our interfaces.
        # Sets up checkpointing, metadata, logging, and optimization-specific state.
        #
        # State within each interface should be encapsulated.
        _BuddyCheckpointing.__init__(self, checkpoint_dir, checkpoint_max_to_keep)
        _BuddyMetadata.__init__(self, metadata_dir)
        _BuddyLogging.__init__(self, log_dir)
        _BuddyOptimizer.__init__(self, optimizer_type, optimizer_checkpoint_interval)

        if optimizer_names is not None:
            warnings.warn(
                "The optimizer_names field is no longer needed, and will be "
                "removed in a future version!",
                DeprecationWarning,
                stacklevel=1,
            )

        # Print available checkpoints
        self._print("Available checkpoint labels:", self.checkpoint_labels)

    def attach_model(self, model: nn.Module) -> None:
        """Attach a model to our Buddy, and move it onto `buddy.device`.

        If a model isn't explicitly passed into the constructor's `model` field,
        `attach_model` should be called before any optimization, checkpointing, etc
        happens.

        Args:
            model (nn.Module): Model to attach.
        """

        assert isinstance(model, nn.Module)

        # Move model to correct device
        model.to(self._device)

        # Attach model
        self._model = model

    # Shared functions
    @property
    def device(self) -> torch.device:
        """Read-only interface for the active torch device. Auto-detected in the
        constructor based on CUDA support.
        """
        return self._device

    def _print(self, *args, **kwargs) -> None:
        """Private helper for logging.
        """
        # Only print in verbose mode
        if self._verbose:
            arglist = list(args)
            arglist[0] = f"[buddy-{self._experiment_name}] {args[0]}"
            print(*arglist, **kwargs)



from typing import Any, List, Tuple, Union, cast


def squeeze(x: Any, axis: Union[int, Tuple[int, ...]] = None) -> Any:
    """Generic squeeze function.

    Args:
        x (Any): Object to squeeze. Must have a `shape` attribute and be indexable with
            slices.
        axis (Union[int, Tuple[int, ...]], optional): Axis or axes to squeeze along.
            If None (default), squeezes all dimensions with value `1`.

    Returns:
        Any: Squeeze object.
    """
    if type(axis) == int:
        axis = cast(Tuple[int, ...], (axis,))
    else:
        axis = cast(Tuple[int, ...], axis)

    slices: List[Union[int, slice]] = []
    for i, dim in enumerate(x.shape):
        if dim == 1 and (axis is None or i in axis):
            slices.append(0)
        elif axis is not None and i in axis:
            assert False, "Desired axis can't be squeezed"
        else:
            slices.append(slice(None))

    return x[tuple(slices)]



from dataclasses import dataclass
from typing import Any

import numpy as np
import pytest
import torch

import fannypack


@dataclass
class Dataclass:
    value: Any
    string: str = "True"
    boolean: bool = False


@pytest.fixture
def nested_conversion():
    X = Dataclass(value={"data": ([np.zeros((10, 1, 5, 1, 10))], np.eye(3))})
    X_torch = fannypack.utils.to_torch(X)
    X_numpy = fannypack.utils.to_numpy(X_torch)
    return X, X_torch, X_numpy


def test_dataclass_checks(nested_conversion):
    """Check for dataclass field issues.
    """
    X, X_torch, X_numpy = nested_conversion

    assert X is not X_torch
    assert X is not X_numpy

    assert X.string == X_torch.string
    assert X.string == X_numpy.string
    assert X.boolean == X_torch.boolean
    assert X.boolean == X_numpy.boolean


def test_conversion_shapes(nested_conversion):
    """Check numpy <=> torch conversions.
    """
    X, X_torch, X_numpy = nested_conversion
    assert X_torch.value["data"][0][0].shape == X_numpy.value["data"][0][0].shape
    assert X_torch.value["data"][1].shape == X_numpy.value["data"][1].shape


def test_conversion_types(nested_conversion):
    """Check numpy <=> torch conversions.
    """
    X, X_torch, X_numpy = nested_conversion
    assert type(X_numpy.value["data"][0][0]) == np.ndarray
    assert type(X_torch.value["data"][0][0]) == torch.Tensor
    assert type(X_numpy.value["data"][1]) == np.ndarray
    assert type(X_torch.value["data"][1]) == torch.Tensor


def test_conversion_values(nested_conversion):
    """Check numpy <=> torch conversions.
    """
    X, X_torch, X_numpy = nested_conversion
    assert np.allclose(X.value["data"][0][0], X_numpy.value["data"][0][0])
    assert np.allclose(X.value["data"][1], X_numpy.value["data"][1])


def test_conversion_failure():
    """Smoke test for an unsupported input type.
    """
    with pytest.raises(AssertionError):
        fannypack.utils.to_torch(None)


def test_to_device():
    """Smoke test for to_device.
    """
    X = {"data": [np.zeros((10, 1, 5, 1, 10))]}
    X_torch = fannypack.utils.to_torch(X)

    X_new = fannypack.utils.to_device(X_torch, torch.device("cpu"), detach=True)
    assert X_torch["data"][0].shape == X_new["data"][0].shape
    assert type(X_new["data"][0]) == torch.Tensor



import abc
from typing import Optional

import torch
import torch.nn as nn


class _BuddyForwardDeclarations(abc.ABC):
    """Abstract class for forward-declaring Buddy's attributes. Enables static
    type-checking in mixins.
    """

    _experiment_name: str
    _model: Optional[nn.Module]
    _device: torch.device

    @abc.abstractproperty
    def device(self) -> torch.device:
        """Read-only interface for the active torch device.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _print(self, *args, **kwargs) -> None:
        """Private helper for logging.
        """
        raise NotImplementedError



from ..data import _trajectories_file
from ._buddy import Buddy
from ._conversions import to_device, to_numpy, to_torch
from ._deprecation import deprecation_wrapper, new_name_wrapper
from ._module_freezing import freeze_module, unfreeze_module
from ._pdb_safety_net import pdb_safety_net
from ._slice_wrapper import SliceWrapper
from ._squeeze import squeeze

DictIterator = new_name_wrapper(
    "fannypack.utils.DictIterator", "fannypack.utils.SliceWrapper", SliceWrapper,
)

TrajectoriesFile = new_name_wrapper(
    "fannypack.utils.TrajectoriesFiles",
    "fannypack.data.TrajectoriesFile",
    _trajectories_file.TrajectoriesFile,
)



from typing import Dict

import torch.nn as nn

_freeze_restore_values: Dict[nn.Module, Dict] = {}


def freeze_module(module: nn.Module, recurse: bool = True) -> None:
    """Freeze the weights of a PyTorch module.

    Args:
        module (torch.nn.Module): Module to freeze.
        recurse (bool, optional): If True, then recursively freezes children.
            Otherwise, only freezes immediate parameters.
    """

    global _freeze_restore_values

    # Recursively call on children
    if recurse:
        for child in module.children():
            freeze_module(child)

    # Do nothing if module is already frozen
    if module in _freeze_restore_values:
        return

    # Freeze parameters
    restore_values = {}
    for name, parameter in module.named_parameters(recurse=False):
        restore_values[name] = parameter.requires_grad
        parameter.requires_grad = False
    _freeze_restore_values[module] = restore_values


def unfreeze_module(module: nn.Module, recurse: bool = True) -> None:
    """Unfreeze the weights of a PyTorch module, which needs to have been
    frozen with :func:`fannypack.utils.freeze_module`.

    Args:
        module (torch.nn.Module): Module to unfreeze.
        recurse (bool, optional): If True, then recursively unfreezes children.
            Otherwise, only unfreezes immediate parameters.
    """

    global _freeze_restore_values

    # Recursively call on children
    if recurse:
        for child in module.children():
            unfreeze_module(child)

    # Do nothing if module is already unfrozen
    if module not in _freeze_restore_values:
        return

    # Freeze parameters
    restore_values = _freeze_restore_values.pop(module)
    for name, parameter in module.named_parameters(recurse=False):
        parameter.requires_grad = restore_values[name]



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="fannypack",
    version="0.0.9",
    description="PyTorch utilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/brentyi/fannypack",
    author="brentyi",
    author_email="brentyi@berkeley.edu",
    license="MIT",
    packages=find_packages(exclude=["examples", "tests"]),
    package_data={"fannypack": ["py.typed"]},
    python_requires=">=3.7",
    install_requires=[
        "dill",
        "h5py",
        "numpy",
        # "numpy-stubs @ https://github.com/numpy/numpy-stubs/tarball/master",
        "pyyaml",
        "tensorboard",
        "torch",
        "tqdm",
        "Pillow",  # Needed for logging images in Tensorboard
        # "pip>=20.1",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from __future__ import annotations

import abc
import glob
import os
import pathlib
import signal
import warnings
from typing import TYPE_CHECKING, Any, Dict, List, cast

import dill
import numpy as np
import torch

from .._forward_declarations import _BuddyForwardDeclarations

if TYPE_CHECKING:
    from ._optimizer import _BuddyOptimizer


class _BuddyCheckpointing(_BuddyForwardDeclarations, abc.ABC):
    """Buddy's model checkpointing interface.
    """

    def __init__(self, checkpoint_dir: str, checkpoint_max_to_keep: int) -> None:
        """Checkpointing-specific setup.
        """
        self._checkpoint_dir = checkpoint_dir
        self._checkpoint_max_to_keep = checkpoint_max_to_keep

        # Find all unlabeled checkpoints for this experiment
        self._checkpoint_unlabeled_files = self._find_unlabeled_checkpoints(
            checkpoint_dir=self._checkpoint_dir, experiment_name=self._experiment_name,
        )

    def save_checkpoint(self, label: str = None, path: str = None) -> None:
        """Saves a checkpoint, which can optionally be labeled.
        """
        assert self._model is not None, "No model attached!"

        # Determine path to checkpoint file
        unlabeled = False
        if path is None and label is None:
            optimizer_steps = cast("_BuddyOptimizer", self).optimizer_steps
            path = "{}/{}-{:016d}.ckpt".format(
                self._checkpoint_dir, self._experiment_name, optimizer_steps,
            )

            if (
                self._checkpoint_unlabeled_files
                and path == self._checkpoint_unlabeled_files[-1]
            ):
                self._print("Skipping redundant checkpoint save")
                return

            unlabeled = True
        elif path is None:
            # Numerical labels are reserved for step counts (see above)
            label = cast(str, label)
            assert not label.isdigit()
            path = "{}/{}-{}.ckpt".format(
                self._checkpoint_dir, self._experiment_name, label
            )

        # Create directory if it doesn't exist yet
        checkpoint_dir = pathlib.Path(path).parents[0]
        if not os.path.isdir(checkpoint_dir):
            os.makedirs(checkpoint_dir)

        # Create state to save. This includes:
        # > Model state
        # > Optimizers
        # > Training steps
        # > Buddy configuration
        optimizer_states = {}

        for name, optimizer in cast("_BuddyOptimizer", self)._optimizer_dict.items():
            optimizer_states[name] = optimizer.state_dict()

        state = {
            "optimizer_config": cast("_BuddyOptimizer", self)._optimizer_config,
            "optimizer_states": optimizer_states,
            "state_dict": self._model.state_dict(),
        }

        # Ignore SIGINT (eg ctrl+c) events while we save to disk...
        try:
            orig_handler = signal.getsignal(signal.SIGINT)
            signal.signal(signal.SIGINT, lambda _sig, _frame: None)
        except ValueError as e:  # pragma: no cover
            # signal throws a ValueError if we're not in the main thread
            self._print("Error while attaching SIGINT handler:", e)
            orig_handler = None

        # "Atomic" checkpoint saving
        tmp_path = "{}/tmp-{}.ckpt".format(checkpoint_dir, np.random.randint(1e10))
        torch.save(state, tmp_path, pickle_module=dill)
        os.rename(tmp_path, path)
        self._print("Saved checkpoint to path:", path)

        # Restore SIGINT handler
        if orig_handler is not None:
            signal.signal(signal.SIGINT, orig_handler)

        # If unlabeled, add to list
        if unlabeled:
            self._checkpoint_unlabeled_files.append(path)

        # Prune checkpoint files
        while len(self._checkpoint_unlabeled_files) > self._checkpoint_max_to_keep:
            os.remove(self._checkpoint_unlabeled_files.pop(0))

    def load_checkpoint_module(
        self,
        source: str,
        target: str = None,
        label: str = None,
        path: str = None,
        experiment_name: str = None,
    ) -> None:
        """Loads parameters from a specific child module within a checkpoint.
        By default, loads the checkpoint with the highest number of training
        iterations.

        Can also be specified via a label or file path.
        """
        assert self._model is not None, "No model attached!"

        if target is None:
            target = source

        # Find and read our checkpoint file
        checkpoint = self._read_checkpoint_file(label, path, experiment_name)

        # Get possible target modules
        module_dict = dict(self._model.named_modules())
        assert target in module_dict.keys(), "Nonexistent target module!"

        # Build a state dict for this module only
        source_state_dict = {}
        key_prefix = ""
        if len(source) > 0:
            key_prefix = f"{source}."
        for key, value in checkpoint["state_dict"].items():
            if key.startswith(key_prefix):
                prefix_length = len(key_prefix)
                source_state_dict[key[prefix_length:]] = value

        # Load state dict
        missing, unexpected = module_dict[target].load_state_dict(source_state_dict)
        assert len(missing) == 0
        assert len(unexpected) == 0

        self._print(f"Loaded module: {source} => {target}")

    def load_checkpoint_optimizer(
        self,
        source: str,
        target: str = None,
        label: str = None,
        path: str = None,
        experiment_name: str = None,
    ) -> None:
        """Loads state associated with a specific optimizer from a checkpoint.
        By default, loads the checkpoint with the highest number of training
        iterations.

        Can also be specified via a label or file path.
        """
        assert self._model is not None, "No model attached!"

        if target is None:
            target = source
        cast("_BuddyOptimizer", self)._instantiate_optimizer(target)

        # Find and read our checkpoint file
        checkpoint = self._read_checkpoint_file(label, path, experiment_name)

        # Input validation
        assert (
            source in checkpoint["optimizer_states"].keys()
        ), "Nonexistent source optimizer!"
        assert (
            target in cast("_BuddyOptimizer", self)._optimizer_dict.keys()
        ), "Nonexistent target optimizer!"

        # Load optimizer state
        state_dict = checkpoint["optimizer_states"][source]
        cast("_BuddyOptimizer", self)._optimizer_dict[target].load_state_dict(
            state_dict
        )
        self._print(f"Loaded optimizer: {source} => {target}")

    def load_checkpoint_optimizers(
        self, label=None, path=None, experiment_name=None
    ) -> None:
        """Loads all optimizer settings from a checkpoint. By default, loads
        the checkpoint with the highest number of training iterations.

        Can also be specified via a label or file path.
        """
        assert self._model is not None, "No model attached!"

        # Find and read our checkpoint file
        checkpoint = self._read_checkpoint_file(label, path, experiment_name)

        # Load optimizer state
        self._load_checkpoint_optimizers(checkpoint)

    def load_checkpoint(
        self, label: str = None, path: str = None, experiment_name: str = None
    ) -> None:
        """Loads a checkpoint. By default, loads the one with the highest
        number of training iterations.

        Can also be specified via a label or file path.
        """
        assert self._model is not None, "No model attached!"

        # Find and read our checkpoint file
        checkpoint = self._read_checkpoint_file(label, path, experiment_name)

        # Load optimizer state
        self._load_checkpoint_optimizers(checkpoint)

        # Load model parameters
        missing, unexpected = self._model.load_state_dict(checkpoint["state_dict"])
        assert len(missing) == 0
        assert len(unexpected) == 0

        optimizer_steps = cast("_BuddyOptimizer", self).optimizer_steps
        self._print("Loaded checkpoint at step:", optimizer_steps)
        return

    @property
    def checkpoint_labels(self) -> List[str]:
        """ Accessor for listing available checkpoint labels.
        These should be saved as: `experiment_name-label.ckpt` in the
        `checkpoint_dir` directory.
        """

        experiment_name = self._experiment_name
        checkpoint_dir = self._checkpoint_dir

        # Find all matching checkpoint files
        path_choices = glob.glob("{}/{}-*.ckpt".format(checkpoint_dir, experiment_name))
        if len(path_choices) == 0:
            return []

        # Pull out labels
        output = []
        for choice in path_choices:
            prefix_len = len(os.path.join(checkpoint_dir, f"{experiment_name}-"))

            suffix_len = len(".ckpt")
            string_label = choice[prefix_len:-suffix_len]
            output.append(string_label)

        # Sort output alphabetically and return
        output.sort()
        return output

    def _load_checkpoint_optimizers(self, checkpoint: Dict[str, Any]) -> None:
        # Load Buddy optimizer configuration
        optimizer_config = cast("_BuddyOptimizer", self)._optimizer_config
        for key, value in checkpoint["optimizer_config"].items():
            if key in optimizer_config.keys():
                optimizer_config[key] = value
            else:
                warnings.warn(f"Skipping missing configuration key: {key}={value}")

        # Instantiate optimizers & load state
        for name, state_dict in checkpoint["optimizer_states"].items():
            cast("_BuddyOptimizer", self)._instantiate_optimizer(name)
            cast("_BuddyOptimizer", self)._optimizer_dict[name].load_state_dict(
                state_dict
            )

    def _read_checkpoint_file(
        self, label: str = None, path: str = None, experiment_name: str = None
    ) -> Dict[str, Any]:
        """Find a checkpoint to load.

        This is one of three options:
          1) The latest, based on step # (must have same experiment name)
          2) A file saved with a label (must have same experiment name)
          3) One specified by a path
        """

        # Determine path to checkpoint file
        if path is None and label is None:
            # Load latest unlabeled checkpoint
            if experiment_name is None:
                # Use our current experiment name by default
                paths = self._checkpoint_unlabeled_files
            else:
                # Use specified experiment name
                paths = self._find_unlabeled_checkpoints(
                    checkpoint_dir=self._checkpoint_dir,
                    experiment_name=experiment_name,
                )
            if len(paths) == 0:
                raise FileNotFoundError("Missing checkpoint file")

            # The list of paths will be sorted by optimizer step count
            path = paths[-1]

        elif path is None and label is not None:
            # Load a labeled checkpoint
            if experiment_name is None:
                # Use our current experiment name by default
                experiment_name = self._experiment_name
            path = os.path.join(self._checkpoint_dir, f"{experiment_name}-{label}.ckpt")
        elif path is not None:
            # Load a checkpoint by its location
            path = path
        else:
            assert (
                False
            ), "Too many arguments! Only one of (label, path) is supported at a time."

        # Load checkpoint dict
        checkpoint = torch.load(path, map_location=self._device, pickle_module=dill)

        # Backwards-compatibility
        # This should eventually be removed :)
        renamed_fields = [
            ("config", "optimizer_config"),
            ("optimizers", "optimizer_states"),
        ]
        for old_name, new_name in renamed_fields:
            if old_name in checkpoint.keys():
                self._print(f"Legacy checkpoint field: {old_name} => {new_name}")
                checkpoint[new_name] = checkpoint[old_name]
                checkpoint.pop(old_name)

        if "steps" in checkpoint.keys():
            self._print("Legacy checkpoint field: steps")
            checkpoint["optimizer_config"]["global_steps"] = checkpoint["steps"]
            checkpoint.pop("steps")

        # Checkpoint file validation
        valid_keys = set(["optimizer_config", "optimizer_states", "state_dict"])
        for key in checkpoint.keys():
            assert key in valid_keys

        # Raise warning for optimizer type mismatches
        if (
            checkpoint["optimizer_config"]["optimizer_type"]
            != cast("_BuddyOptimizer", self)._optimizer_config["optimizer_type"]
        ):
            warnings.warn("Checkpoint loading: overriding optimizer type.")

        self._print("Read checkpoint from path:", path)
        return checkpoint

    @staticmethod
    def _find_unlabeled_checkpoints(
        checkpoint_dir: str, experiment_name: str
    ) -> List[str]:
        """(Private) Returns a list of all unlabeled checkpoints associated
        with this experiment, sorted from oldest to newest.
        """

        # Find all matching checkpoint files
        path_choices = glob.glob("{}/{}-*.ckpt".format(checkpoint_dir, experiment_name))
        if len(path_choices) == 0:
            return []

        # Find unlabeled checkpoint files + associated step counts
        output = []
        for choice in path_choices:
            prefix_len = len("{}/{}-".format(checkpoint_dir, experiment_name))

            suffix_len = len(".ckpt")
            string_steps = choice[prefix_len:-suffix_len]
            try:
                steps = int(string_steps)
                output.append((choice, steps))
            except ValueError:
                pass

        # Sort output by steps
        output.sort(key=lambda x: x[1])

        # Return paths only
        return [x[0] for x in output]



import os

import numpy as np
import pytest
import torch
import torch.nn.functional as F

import fannypack

from ..fixtures import resblock_buddy_temporary_data, simple_buddy


def test_buddy_metadata(resblock_buddy_temporary_data):
    """Do some checks on rapid continuous checkpointing.
    """
    model, buddy, data, labels = resblock_buddy_temporary_data

    # Check initial metadata state
    assert buddy.metadata == {}

    buddy.set_metadata({"number": 5})
    buddy.add_metadata({"string": "words words"})

    # Load from file that was saved
    buddy.load_metadata()

    # Check that contents are valid
    assert len(buddy.metadata) == 2
    assert buddy.metadata["number"] == 5

    # Load from file that was saved
    buddy.load_metadata(path=buddy.metadata_path)

    # Check that contents are valid
    assert len(buddy.metadata) == 2
    assert buddy.metadata["number"] == 5



import os
import signal

import requests
from tqdm.auto import tqdm

_cache_path = os.path.expanduser("~/.cache/fannypack-drive-files")


def set_cache_path(path: str):
    """Set the cache location for :func:`fannypack.data.cached_drive_file`.

    Args:
        _cache_path (str): New location for cached files. Defaults to
            `~/.cache/fannypack-drive-files/`.
    """
    global _cache_path
    _cache_path = path


def cached_drive_file(name: str, url: str) -> str:
    """Return a local path to a file from Google Drive. Downloads the file if it doesn't
    exist yet locally.

    By default, cached files live in `~/.cache/fannypack-drive-files/`. It often makes
    sense to move this directory (eg to an NFS): see
    :func:`fannypack.data.set_cache_path`.

    Args:
        name (str): Name of path, eg `secret_key.pem`.
        url (str): URL, eg
            `https://drive.google.com/file/d/1AsY9Cs3xE0RSlr0FKlnSKHp6zIwFSvXe/view`.

    Returns:
        str: Local path to file.
    """

    # Generated cached filename
    cached_filename = f"{_drive_id_from_url(url)}-{name}"
    cached_filepath = os.path.join(_cache_path, cached_filename)

    if not os.path.exists(cached_filepath):
        print(f"[fannypack-drive] Downloading file to {cached_filepath}")
        download_drive_file(url, cached_filepath)
        assert os.path.exists(cached_filepath)
    return cached_filepath


def download_drive_file(url: str, target_path: str, chunk_size=32768) -> None:
    """Download a file via a public Google Drive url.

    Example usage:
    ```
    download_file_from_google_drive(
        "https://drive.google.com/file/d/1AsY9Cs3xE0RSlr0FKlnSKHp6zIwFSvXe/view",
        "/home/brent/Downloads/test.pdf"
    )
    ```

    Args:
        url (str): Google Drive url.
        target_path (str): Destination to write to.
    """
    # Create directory if it doesn't exist yet
    directory = os.path.dirname(target_path)
    if not os.path.isdir(directory):
        os.makedirs(directory)
        print("[fannypack-drive] Created directory:", directory)

    # Parse URL
    drive_id = _drive_id_from_url(url)
    download_url = "https://docs.google.com/uc?export=download"

    # Download file
    session = requests.Session()
    response = session.get(
        download_url,
        params={"id": drive_id},
        stream=True,
        headers={"Accept-Encoding": None},
    )

    token = None
    for key, value in response.cookies.items():
        if key.startswith("download_warning"):  # pragma: no cover
            params = {"id": drive_id, "confirm": value}
            response = session.get(download_url, params=params, stream=True)
            break

    # Delete partially downloaded files if we hit interrupt (Ctrl+C) before download
    # finishes
    try:
        orig_handler = signal.getsignal(signal.SIGINT)

        def sigint_handler(sig, frame):  # pragma: no cover
            print("[fannypack-drive] Deleting file:", target_path)
            os.remove(target_path)
            orig_handler(sig, frame)
            # Restore SIGINT handler
            if orig_handler is not None:
                signal.signal(signal.SIGINT, orig_handler)

        signal.signal(signal.SIGINT, sigint_handler)
    except ValueError as e:  # pragma: no cover
        # signal throws a ValueError if we're not in the main thread
        print("[fannypack-drive] Error while attaching SIGINT handler:", e)
        orig_handler = None

    # Download file
    progress_bar = tqdm(unit="iB", unit_scale=True)
    with open(target_path, "wb") as f:
        for chunk in response.iter_content(chunk_size):
            # Filter out keep-alive new chunks
            if chunk:
                progress_bar.update(len(chunk))
                f.write(chunk)

    # Restore SIGINT handler
    if orig_handler is not None:
        signal.signal(signal.SIGINT, orig_handler)


def _drive_id_from_url(url: str) -> str:
    """Get an ID from a Google Drive URL.

    Args:
        url (str): Publicly visible Google Drive URL.

    Returns:
        str: Google Drive ID.
    """

    url_prefixes = [
        "https://drive.google.com/file/d/",
        "https://drive.google.com/open?id=",
    ]

    for prefix in url_prefixes:
        if url.startswith(prefix):
            parts = url[len(prefix) :].split("/")
            assert len(parts) in (1, 2)
            return parts[0]
    assert False, f"Malformed Google Drive URL: {url}"
